/******************************************
 *    BeOS Infra-Red Remote Control       *
 *             Serial IR driver           *
 *                                        *
 * This is the driver for serial IR       *
 * receivers/transmitters                 *
 ******************************************/
#include <drivers/Drivers.h>
#include <drivers/config_manager.h>
#include <drivers/isapnp.h>
#include <drivers/module.h>
#include <drivers/ISA.h>

#include <sys/time.h>

#include "serial.h"
#include "lirc.h"

/*********************
 *    protoypes      *
 *********************/
status_t open_hook(const char *name, uint32 flags, void **cookie);
status_t close_hook(void *cookie);
status_t free_hook(void *cookie);
status_t read_hook(void *cookie, off_t position, void *data, size_t *len);
status_t write_hook(void *cookie, off_t position, const void *data, size_t *len);
status_t control_hook(void *cookie, uint32 op, void *data, size_t len);

status_t init_port(TS_SERIAL_device *ps_Device);

/***************************
 *    Global variables     *
 ***************************/
TS_SERIAL_device  gas_serial_Devices[SERIAL_DEVICE_MAX];
static isa_module_info  *gps_Isa;

int type = LIRC_HOMEBREW;  /* default type. TODO: IOCTL to change it */

TS_SERIAL_device hardware[]=  /* init values for each type */
{
   /* home-brew receiver/transmitter */
   {
      true, 0, 0, "", 0, 0, -1,
      LIRC_HOMEBREW,
      UART_MSR_DCD,
      UART_MSR_DDCD,
      UART_MCR_RTS|UART_MCR_OUT2|UART_MCR_DTR,
      UART_MCR_RTS|UART_MCR_OUT2,
      (
#ifdef LIRC_SERIAL_TRANSMITTER
       LIRC_CAN_SET_SEND_DUTY_CYCLE|
       LIRC_CAN_SET_SEND_CARRIER|
       LIRC_CAN_SEND_PULSE|
#endif
       LIRC_CAN_REC_MODE2)
   },

   /* IRdeo classic */
   {
      true, 0, 0, "", 0, 0, -1,
      LIRC_IRDEO,
      UART_MSR_DSR,
      UART_MSR_DDSR,
      UART_MCR_OUT2,
      UART_MCR_RTS|UART_MCR_DTR|UART_MCR_OUT2,
      (LIRC_CAN_SET_SEND_DUTY_CYCLE|
       LIRC_CAN_SEND_PULSE|
       LIRC_CAN_REC_MODE2)
   },

   /* IRdeo remote */
   {
      true, 0, 0, "", 0, 0, -1,
      LIRC_IRDEO_REMOTE,
      UART_MSR_DSR,
      UART_MSR_DDSR,
      UART_MCR_RTS|UART_MCR_DTR|UART_MCR_OUT2,
      UART_MCR_RTS|UART_MCR_DTR|UART_MCR_OUT2,
      (LIRC_CAN_SET_SEND_DUTY_CYCLE|
       LIRC_CAN_SEND_PULSE|
       LIRC_CAN_REC_MODE2)
   },

   /* AnimaX */
   {
      true, 0, 0, "", 0, 0, -1,
      LIRC_ANIMAX,
      UART_MSR_DCD,
      UART_MSR_DDCD,
      0,
      UART_MCR_RTS|UART_MCR_DTR|UART_MCR_OUT2,
      LIRC_CAN_REC_MODE2
   }
};

#define RS_ISR_PASS_LIMIT 256

/* A long pulse code from a remote might take upto 300 bytes.  The
   daemon should read the bytes as soon as they are generated, so take
   the number of keys you think you can push before the daemon runs
   and multiply by 300.  The driver will warn you if you overrun this
   buffer.  If you have a slow computer or non-busmastering IDE disks,
   maybe you will need to increase this.  */

/* This MUST be a power of two!  It has to be larger than 1 as well. */

#define RBUF_LEN 256

static lirc_t rbuf[RBUF_LEN];
static int rbh, rbt;

static struct timeval lasttv = {0, 0};

int32 api_version = B_CUR_DRIVER_API_VERSION;

/******************************
 *    Port access functions   *
 ******************************/
static inline unsigned int sinp(uint32 addr, int offset)
{
   /* return (*gps_Isa->read_io_8(addr + offset)); */
   return (gps_Isa->read_io_8(addr + offset));
   /* return (read_io_8(addr + offset)); */
}

static inline void soutp(uint32 addr, int offset, int value)
{
   /* *gps_Isa->write_io_8(value, addr + offset); */
   gps_Isa->write_io_8(addr + offset, value);
   /* write_io_8(value, addr + offset); */
}

void on(TS_SERIAL_device *ps_Device)
{
   soutp(ps_Device->base_addr, UART_MCR, ps_Device->on);
}

void off(TS_SERIAL_device *ps_Device)
{
   soutp(ps_Device->base_addr, UART_MCR, ps_Device->off);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 29/04/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Called when the system is booted, to let the driver      *
 * detect and reset the hardware.                           *
 ************************************************************/
status_t init_hardware(void)
{
   DBG_PRINT("init_hardware");
   return(B_OK);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 29/04/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Called when the driver is loaded, so it can allocate     *
 * needed system resources.                                 *
 ************************************************************/
status_t init_driver(void)
{
   uint8 i;
   status_t i_ret;
   status_t s = get_module( B_ISA_MODULE_NAME, (module_info **) &gps_Isa);
   if(s != B_OK)
      return(s);
   
   DBG_PRINT("init_driver");
   i_ret = SERIAL_InitDeviceArray(gas_serial_Devices, &(hardware[type]));
   if(i_ret != B_OK)
   {
      DBG_PRINT("init_driver: InitDeviceArray error");
      return(i_ret);
   }

   /* find all configured serial ports */
   detect(gas_serial_Devices);

   /* TODO: drivers settings API to set the type of each device 
      connected on each port */
   /* initialise ports */
   for(i=0; i<SERIAL_DEVICE_MAX; i++)
   {
      if(gas_serial_Devices[i].is_free == false);
      {
         i_ret = init_port(&(gas_serial_Devices[i]));
         if(i_ret != B_OK)
         {
            DBG_PRINT("init_driver: init_port error");
            return(i_ret);            
         }
      }
   }
   return(B_OK);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 29/04/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Called just before the driver is unloaded, so it can     *
 * free allocated resources.                                *
 ************************************************************/
void uninit_driver(void)
{
   DBG_PRINT("uninit_driver");
   SERIAL_FiniDeviceArray(gas_serial_Devices);
   put_module( B_ISA_MODULE_NAME);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 29/04/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Called to obtain a list of device names supported by the *
 * driver.                                                  *
 ************************************************************/
const char **publish_devices(void)
{
   const static char *gas_DeviceList[SERIAL_DEVICE_MAX + 1];
   uint16 i;
   uint16 j = 0;
   
   DBG_PRINT("publish_devices");

   for (i=0; i<SERIAL_DEVICE_MAX; ++i)
   {
      if (gas_serial_Devices[i].is_free == false)
      {
         gas_DeviceList[j++] = gas_serial_Devices[i].name;
         DBG_PRINT(gas_serial_Devices[i].name);
      }
   }
   gas_DeviceList[j] = 0;
   return (gas_DeviceList);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 29/04/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Called to obtain a list of pointers to the hook          *
 * functions for a specified device.                        *
 ************************************************************/
device_hooks *find_device(const char *name)
{
   static device_hooks dh = 
   {
      open_hook, 
      close_hook, 
      free_hook,
      control_hook, 
      read_hook, 
      write_hook,
      NULL, NULL, NULL, NULL
   };
   DBG_PRINT("find_device");
   return (&dh);
}


static void inline rbwrite(lirc_t l)
{
   unsigned int nrbt;

   nrbt=(rbt+1) & (RBUF_LEN-1);
   if(nrbt==rbh)      /* no new signals will be accepted */
   {
#ifdef DEBUG
      printk(KERN_WARNING  LIRC_DRIVER_NAME  ": Buffer overrun\n");
#endif
      return;
   }
   rbuf[rbt]=l;
   rbt=nrbt;
}

static void inline frbwrite(lirc_t l)
{
   /* simple noise filter */
   static lirc_t pulse=0L,space=0L;
   static unsigned int ptr=0;

   if(ptr>0 && (l&PULSE_BIT))
   {
      pulse+=l&PULSE_MASK;
      if(pulse>250)
      {
         rbwrite(space);
         rbwrite(pulse|PULSE_BIT);
         ptr=0;
         pulse=0;
      }
      return;
   }
   if(!(l&PULSE_BIT))
   {
      if(ptr==0)
      {
         if(l>20000)
         {
            space=l;
            ptr++;
            return;
         }
      }
      else
      {
         if(l>20000)
         {
            space+=pulse;
            if(space>PULSE_MASK) space=PULSE_MASK;
            space+=l;
            if(space>PULSE_MASK) space=PULSE_MASK;
            pulse=0;
            return;
         }
         rbwrite(space);
         rbwrite(pulse|PULSE_BIT);
         ptr=0;
         pulse=0;
      }
   }
   rbwrite(l);
}

int32 irq_handler(void *handler_data)
{
   struct timeval tv;
   int status,counter,dcd;
   int prev_rbt;
   long deltv;
   lirc_t data;
   TS_SERIAL_device *ps_Device;

   ps_Device = (TS_SERIAL_device*)handler_data;
   
   counter=0;
   do
   {
      counter++;
      status=sinp(ps_Device->base_addr, UART_MSR);
      if(counter>RS_ISR_PASS_LIMIT)
      {
         DBG_PRINT("AIEEEE: We're caught!\n");
         break;
      }
      if((status &  ps_Device->signal_pin_change) && ps_Device->sense!=-1)
      {
         /* get current time */
         gettimeofday(&tv, NULL);

         /* New mode, written by Trent Piepho 
            <xyzzy@u.washington.edu>. */

         /* The old format was not very portable.
            We now use the type lirc_t to pass pulses
            and spaces to user space.

            If PULSE_BIT is set a pulse has been
            received, otherwise a space has been
            received.  The driver needs to know if your
            receiver is active high or active low, or
            the space/pulse sense could be
            inverted. The bits denoted by PULSE_MASK are
            the length in microseconds. Lengths greater
            than or equal to 16 seconds are clamped to
            PULSE_MASK.  All other bits are unused.
            This is a much simpler interface for user
            programs, as well as eliminating "out of
            phase" errors with space/pulse
            autodetection. */

         /* calculate time since last interrupt in
            microseconds */
         dcd = (status & ps_Device->signal_pin) ? 1:0;

         deltv=tv.tv_sec-lasttv.tv_sec;
         if(deltv>15) 
         {
#ifdef DEBUG
            DBG_PRINT("AIEEEE: %d %d %lx %lx %lx %lx\n",
                     dcd,ps_Device->sense,
                     tv.tv_sec,lasttv.tv_sec,
                     tv.tv_usec,lasttv.tv_usec);
#endif
            data=PULSE_MASK; /* really long time */
            if(!(dcd^ps_Device->sense)) /* sanity check */
            {
               /* detecting pulse while this
                  MUST be a space! */
               ps_Device->sense= ps_Device->sense ? 0:1;
            }
         }
         else
         {
            data=(lirc_t) (deltv*1000000+
                  tv.tv_usec-
                  lasttv.tv_usec);
         };
         if(tv.tv_sec<lasttv.tv_sec ||
               (tv.tv_sec==lasttv.tv_sec &&
                tv.tv_usec<lasttv.tv_usec))
         {
            DBG_PRINT("AIEEEE: your clock just jumped backwards\n");
            data=PULSE_MASK;
         }
         prev_rbt = rbt;
         frbwrite(dcd^ps_Device->sense ? data : (data|PULSE_BIT));

         if(prev_rbt != rbt)
         {
            if( !(ps_Device->flags & O_NONBLOCK))
               release_sem_etc(ps_Device->read_sem, 1, B_DO_NOT_RESCHEDULE);
            if(ps_Device->is_waiting_data == true)
               release_sem_etc(ps_Device->wait_data_sem, 1, B_DO_NOT_RESCHEDULE);
         }
         lasttv=tv;
      }
   } 
   while(!(sinp(ps_Device->base_addr, UART_IIR) & UART_IIR_NO_INT)); /* still pending ? */

   return(B_HANDLED_INTERRUPT);
}

status_t init_port(TS_SERIAL_device *ps_Device)
{
   cpu_status s_CpuStatus;
   
   /* Check io region : how to do this on BeOS ??*/

   s_CpuStatus = disable_interrupts();

   dprintf("BERC - SERIAL : init_port (addr 0x%X irq %d)\n",
            (int)ps_Device->base_addr, (int)ps_Device->irq);

   /* Set DLAB 0. */
   soutp(ps_Device->base_addr, UART_LCR, 
         sinp(ps_Device->base_addr, UART_LCR) & (~UART_LCR_DLAB));

   /* First of all, disable all interrupts */
   soutp(ps_Device->base_addr, UART_IER, 
         sinp(ps_Device->base_addr, UART_IER)&
         (~(UART_IER_MSI|UART_IER_RLSI|UART_IER_THRI|UART_IER_RDI)));

   /* Clear registers. */
   sinp(ps_Device->base_addr, UART_LSR);
   sinp(ps_Device->base_addr, UART_RX);
   sinp(ps_Device->base_addr, UART_IIR);
   sinp(ps_Device->base_addr, UART_MSR);

   /* Set line for power source */
   soutp(ps_Device->base_addr, UART_MCR, ps_Device->off);

   /* Clear registers again to be sure. */
   sinp(ps_Device->base_addr, UART_LSR);
   sinp(ps_Device->base_addr, UART_RX);
   sinp(ps_Device->base_addr, UART_IIR);
   sinp(ps_Device->base_addr, UART_MSR);

   switch(ps_Device->type)
   {
      case LIRC_IRDEO:
      case LIRC_IRDEO_REMOTE:
         /* setup port to 7N1 @ 115200 Baud */
         /* 7N1+start = 9 bits at 115200 ~ 3 bits at 38kHz */

         /* Set DLAB 1. */
         soutp(ps_Device->base_addr, 
               UART_LCR, sinp(ps_Device->base_addr, UART_LCR) | UART_LCR_DLAB);
         /* Set divisor to 1 => 115200 Baud */
         soutp(ps_Device->base_addr, UART_DLM,0);
         soutp(ps_Device->base_addr, UART_DLL,1);
         /* Set DLAB 0 +  7N1 */
         soutp(ps_Device->base_addr, UART_LCR,UART_LCR_WLEN7);
         /* THR interrupt already disabled at this point */
         break;
      default:
         break;
   }

   restore_interrupts(s_CpuStatus);
   
#ifdef USE_RDTSC
   /* Initialize pulse/space widths */
   calc_pulse_lengths_in_clocks();
#endif

   /* If pin is high, then this must be an active low receiver. */
   if(ps_Device->sense==-1)
   {
      snooze(100000); /* wait 100 ms */
      ps_Device->sense=(sinp(ps_Device->base_addr, UART_MSR) & ps_Device->signal_pin) ? 1:0;
      if(ps_Device->sense)
         DBG_PRINT("auto-detected active low receiver\n");
      else
         DBG_PRINT("auto-detected active high receiver\n");
   }
   else
   {
      if(ps_Device->sense)
         DBG_PRINT("Manually using active low receiver\n");
      else
         DBG_PRINT("Manually using active high receiver\n");
   };

   return(B_OK);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 03/06/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Open function                                            *
 ************************************************************/
status_t open_hook(const char *name, uint32 flags, void **cookie)
{
   uint8 i;
   long i_ret;
   cpu_status s_CpuStatus;
   TS_SERIAL_device *ps_Device;

#warning TODO: report spinlock used on lirc?
   /* find device and check that it is not already open */
   for(i=0; i<SERIAL_DEVICE_MAX; i++)
   {
      dprintf("BERC - SERIAL : Open check %s\n", gas_serial_Devices[i].name);

      if(strcmp(gas_serial_Devices[i].name, name) == 0)
      {
         if( (gas_serial_Devices[i].is_free == false) &&
             (gas_serial_Devices[i].nb_open == 0) )
         {
            ps_Device = &(gas_serial_Devices[i]);
            ps_Device->nb_open++;
            break;
         }
         else
         {
            dprintf("BERC - SERIAL : Unable to open device %s!\n", name);
            return(B_NAME_NOT_FOUND);
         }             
      }
   }
   if(i == SERIAL_DEVICE_MAX)
   {
      dprintf("BERC - SERIAL : device %s not found!\n", name);
      return(B_NAME_NOT_FOUND);
   }
   
   dprintf("BERC - SERIAL : open using descriptor %d (addr 0x%X irq %d)\n", i,
            (int)ps_Device->base_addr, (int)ps_Device->irq);
   
   /* init cookie */
   *cookie = ps_Device;
   ps_Device->flags = flags;
   
   /* initialize timestamp */
   gettimeofday(&lasttv, NULL);

   i_ret = install_io_interrupt_handler(ps_Device->irq, irq_handler, ps_Device, 0);
   if(i_ret != B_OK)
   {
      DBG_PRINT("Error while installing irq handler");
   }

   /* finally enable interrupts. */
   s_CpuStatus = disable_interrupts();

   /* Set DLAB 0. */
   soutp(ps_Device->base_addr, UART_LCR, 
         sinp(ps_Device->base_addr, UART_LCR) & (~UART_LCR_DLAB));   

   soutp(ps_Device->base_addr, UART_IER, 
         sinp(ps_Device->base_addr, UART_IER)|UART_IER_MSI);

   /* Init read buffer pointers. */
   rbh = rbt = 0;
   restore_interrupts(s_CpuStatus);
   return 0;
}

/************************************************************
 *  Function :                                              *
 *  Date     : 03/06/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Close function                                           *
 ************************************************************/
status_t close_hook(void *cookie)
{
   cpu_status s_CpuStatus;
   TS_SERIAL_device *ps_Device;

   ps_Device = cookie;
   s_CpuStatus = disable_interrupts();

   /* unlock current userland calls */
   if( (ps_Device->is_reading == true) &&
       !(ps_Device->flags & O_NONBLOCK))
   {
      release_sem(ps_Device->read_sem);
   }
   if(ps_Device->is_waiting_data == true)
      release_sem(ps_Device->wait_data_sem);
   
   /* Set DLAB 0. */
   soutp(ps_Device->base_addr, UART_LCR, 
         sinp(ps_Device->base_addr, UART_LCR) & (~UART_LCR_DLAB));

   /* First of all, disable all interrupts */
   soutp(ps_Device->base_addr, UART_IER, 
         sinp(ps_Device->base_addr, UART_IER)&
         (~(UART_IER_MSI|UART_IER_RLSI|UART_IER_THRI|UART_IER_RDI)));
   restore_interrupts(s_CpuStatus);

   remove_io_interrupt_handler(ps_Device->irq, irq_handler, ps_Device);
   ps_Device->nb_open--;
   return(B_OK);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 03/06/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * free  function                                           *
 ************************************************************/
status_t free_hook(void *cookie)
{
#warning TODO: report parts of the close here?
   return(B_OK);
}
/************************************************************
 *  Function :                                              *
 *  Date     : 03/06/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Read  function                                           *
 ************************************************************/
status_t read_hook(void *cookie, off_t position, void *data, size_t *len)
{
   int n=0;
   uint8 *buf = data;
   TS_SERIAL_device *ps_Device = cookie;

   ps_Device->is_reading = true;
   while (n < *len)
   {
      if( !(ps_Device->flags & O_NONBLOCK))
         acquire_sem(ps_Device->read_sem);
      if (rbt != rbh) 
      {
         memcpy( (void *) buf + n, (void *) &rbuf[rbh], sizeof(lirc_t));
         rbh = (rbh + 1) & (RBUF_LEN - 1);
         n += sizeof(lirc_t);
      } 
      else
      {
         if (ps_Device->flags & O_NONBLOCK) 
            break;
      }
   }
   *len = n;
   ps_Device->is_reading = false;
   return (B_OK);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 03/06/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * Write function; Just a stub for the moment.              *
 ************************************************************/
status_t write_hook(void *cookie, off_t position, const void *data, size_t *len)
{
   return(B_PERMISSION_DENIED);
}

/************************************************************
 *  Function :                                              *
 *  Date     : 03/06/2003                                   *
 ************************************************************
 * Usage:                                                   *
 * ioctl function                                           *
 ************************************************************/
status_t control_hook(void *cookie, uint32 op, void *data, size_t len)
{
   uint32   *p_Data = (uint32*)data;
   TS_SERIAL_device *ps_Device = cookie;

   switch(op)
   {
      /* BeOS base ioctls */
      case B_SET_NONBLOCKING_IO:
         ps_Device->flags |=  O_NONBLOCK;
         break;
      
      case B_SET_BLOCKING_IO:
         ps_Device->flags &=  (~O_NONBLOCK);
         break;

      case B_GET_READ_STATUS:
         if (rbt != rbh) 
            *p_Data = 1;
         else
            *p_Data = 0;
         break;

      case B_GET_WRITE_STATUS:
         *p_Data = 0;
         break;

      /* lirc ioctls */
      case LIRC_WAIT_DATA:    /* Wait for input data */         
         if (rbt != rbh)      /* data already available */
         {
            *p_Data = 1;
            break;
         }

         /* first empty sem count */
         while(acquire_sem_etc(ps_Device->wait_data_sem, 1, B_RELATIVE_TIMEOUT, 0) == B_OK);

         /* then wait for data */
         ps_Device->is_waiting_data = true;
         if (rbt == rbh) 
         {
            acquire_sem_etc(ps_Device->wait_data_sem, 1, B_RELATIVE_TIMEOUT, *p_Data);
            if (rbt != rbh)
               *p_Data = 1;
            else
               *p_Data = 0;
         }
         else  /* data arrived between first test and the flag set */
            *p_Data = 1;
         ps_Device->is_waiting_data = false;
         break;

      case LIRC_GET_FEATURES:
         *p_Data = ps_Device->features;
         break;

      case LIRC_GET_SEND_MODE:
         if(!(ps_Device->features&LIRC_CAN_SEND_MASK))
            return(B_DEV_INVALID_IOCTL);
         
         *p_Data = LIRC_SEND2MODE(ps_Device->features&LIRC_CAN_SEND_MASK);
         break;

      case LIRC_GET_REC_MODE:
         if(!(ps_Device->features&LIRC_CAN_REC_MASK))
            return(B_DEV_INVALID_IOCTL);
         
         *p_Data = LIRC_REC2MODE(ps_Device->features&LIRC_CAN_REC_MASK);

      case LIRC_SET_REC_MODE:
         if(!(ps_Device->features&LIRC_CAN_REC_MASK))
            return(B_DEV_INVALID_IOCTL);
         if((*(uint32*)data) != LIRC_MODE_MODE2) 
            return(B_DEV_INVALID_IOCTL);

      case LIRC_SET_SEND_DUTY_CYCLE:
      case LIRC_SET_SEND_CARRIER:
      default:
         return(B_DEV_INVALID_IOCTL);
   }
   return(B_OK);
}

